#     retry_pytest is a library for pytest test framework
#     Copyright (C) 2021  Alexander Evdokimov
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import ceil
from time import sleep
from typing import Callable, List

import allure
from allure_commons import plugin_manager
from allure_commons.utils import uuid4

from retry_pytest.command import Command
from retry_pytest.custom_errors import NegativeAction


class Retry:
    def __init__(
            self, *expected_exceptions, timeout=10, poll_frequency=1,
            title='Retry', error_msg='', show_expected=False, **kwargs):
        self._exceptions = expected_exceptions if expected_exceptions else (NegativeAction,)
        self._timeout = timeout
        error_msg = error_msg if error_msg else f'timeout {timeout}s exceeded'
        self._timeout_msg = f'{title}: {error_msg}'
        self._poll_frequency = poll_frequency
        self._command_queue = []
        self._title = title
        self._step_params = kwargs
        self._current_step = None
        self._show_expected = show_expected

    @property
    def commands(self) -> List[Command]:
        return self._command_queue

    @property
    def last_command(self) -> Command:
        return self._command_queue[-1]

    def __enter__(self):
        self._current_step = uuid4()
        plugin_manager.hook.start_step(uuid=self._current_step, title=self._title, params=self._step_params)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            for _ in range(ceil(self._timeout / self._poll_frequency)):
                try:
                    sleep(self._poll_frequency)
                    if all([f() for f in self._command_queue]):
                        break
                except self._exceptions as e:
                    if self._show_expected:
                        allure.attach(
                            f'{e.__class__.__name__}: {str(e)}', 'Expected exception', allure.attachment_type.TEXT
                        )
            else:
                exc_type = AssertionError
                exc_val = exc_type(self._timeout_msg)
                exc_tb = None
        except Exception as e:
            exc_type = e.__class__
            exc_val = e
            exc_tb = None
        plugin_manager.hook.stop_step(
            uuid=self._current_step,
            title=self._title,
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb
        )
        if exc_val:
            raise exc_val

    def check(self, func: Callable, *args, **kwargs) -> Command:
        self._command_queue.append(Command(func, *args, **kwargs))
        return self._command_queue[-1]



#     retry_pytest is a library for pytest test framework
#     Copyright (C) 2021  Alexander Evdokimov
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


class NegativeAction(Exception):
    pass


class RetryActionError(Exception):
    pass



#     retry_pytest is a library for pytest test framework
#     Copyright (C) 2021  Alexander Evdokimov
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Callable


class ResultHandler(ABC):
    @abstractmethod
    def set_next(self, handler: ResultHandler) -> ResultHandler:
        pass

    @abstractmethod
    def handle(self, result, cmd):
        pass


class BaseHandler(ResultHandler):
    def __init__(self):
        self._next = None

    def set_next(self, handler: ResultHandler) -> ResultHandler:
        self._next = handler
        return handler

    @abstractmethod
    def handle(self, result, item):
        if self._next:
            return self._next.handle(result, item)


class CallableHandler(BaseHandler):
    def handle(self, result, item):
        if callable(item):
            return item(result)
        return super().handle(result, item)


class ListHandler(BaseHandler):
    def handle(self, result, item):
        if isinstance(result, (list, tuple)):
            return result[int(item)]
        return super().handle(result, item)


class DictHandler(BaseHandler):
    def handle(self, result, item):
        if isinstance(result, dict):
            return result.get(item)
        return super().handle(result, item)


class ObjectHandler(BaseHandler):
    def handle(self, result, item):
        attr = getattr(result, item)
        if isinstance(attr, property):
            return attr
        if callable(attr):
            return attr()
        return attr


class Command:
    def __init__(self, func: Callable, *args, **kwargs):
        self._condition = lambda: False
        self._result_value = None
        self._result_history = []
        self._command = func
        self._args = args
        self._kwargs = kwargs
        self._actions = []
        handler = self._action_handler = CallableHandler()
        for h in [ListHandler(), DictHandler(), ObjectHandler()]:
            handler = handler.set_next(h)

    @property
    def result_history(self):
        return self._result_history

    @property
    def result(self):
        self._result_history.clear()
        self._result_history.append(self._result_value)
        if self._actions:
            for item in self._actions:
                self._result_history.append(
                    self._action_handler.handle(self._result_history[-1], item)
                )
        return self._result_history[-1]

    def less(self, other):
        self._condition = lambda: self.result < other

    def less_equal(self, other):
        self._condition = lambda: self.result <= other

    def equal(self, other):
        self._condition = lambda: self.result == other

    def not_equal(self, other):
        self._condition = lambda: self.result != other

    def greater(self, other):
        self._condition = lambda: self.result > other

    def greater_equal(self, other):
        self._condition = lambda: self.result >= other

    def len(self):
        self._actions.append(len)
        return self

    def is_(self, other):
        self._condition = lambda: self.result is other

    def is_not(self, other):
        self._condition = lambda: self.result is not other

    def in_(self, other):
        self._condition = lambda: self.result in other

    def not_in(self, other):
        self._condition = lambda: self.result not in other

    def get(self, attribute_name=None):
        self._actions.append(attribute_name)
        return self

    def contains(self, data):
        self._condition = data in self.result

    def __call__(self, *args, **kwargs):
        self._result_value = self._command(*self._args, **self._kwargs)
        return self._condition()



import os
import setuptools

from sphinx.setup_command import BuildDoc

import retry_pytest

with open('README.md', 'r') as fh:
    long_description = fh.read()

with open('install_requires.txt', 'r') as ir:
    install_requires = [item.strip('\n') for item in ir.readlines()]

version = os.environ.get('PKG_VERSION') or retry_pytest.__version__

dependency_links = []

setuptools.setup(
    name=retry_pytest.__name__.lower(),
    version=version,
    author=retry_pytest.__author__,
    author_email=retry_pytest.__author_email__,
    description='retry_pytest',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(exclude=['tests']),
    include_package_data=True,
    python_requires='>=3.7',
    install_requires=install_requires,
    dependency_links=dependency_links,
    url='https://github.com/Shumodan/retry-pytest',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Testing',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: OS Independent',
        'Framework :: Pytest',
    ],
    cmdclass={'build_doc': BuildDoc},
    command_options={
        'build_doc': {
            'project': ('setup.py', retry_pytest.__name__),
            'version': ('setup.py', version),
            'source_dir': ('setup.py', os.path.join(retry_pytest.__name__, 'docs', 'source')),
            'build_dir': ('setup.py', os.path.join(retry_pytest.__name__, 'docs', 'build'))
        }
    },
    package_data={
        f'{retry_pytest.__name__}.docs.build': ['html/*', 'html/*/*'],
    },
)
